#include <QColor>
#include <QSlider>
#include<QSplitter>
#include <QLabel>
#include<QPlainTextEdit>

#include <QVBoxLayout>
#include <QHBoxLayout>
#include<QLineEdit>
#include<QSpinBox>
#include<QComboBox>

#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QHeaderView> 

#include "rviz/visualization_manager.h"
#include "rviz/render_panel.h"
#include "rviz/display.h"

#include "myviz.h"
#include "../tools/color.h"

// BEGIN_TUTORIAL
// Constructor for MyViz.  This does most of the work of the class.
MyViz::MyViz( QWidget* parent)
  : QWidget( parent )
{

  nh=ros::NodeHandle("~");
  sub=nh.subscribe("/test",1000,&MyViz::subCallback,this);
  pub=nh.advertise<std_msgs::String>("/test_pub",1000);
  pub_thread=std::thread(&MyViz::pubThread,this);

  // 全局变量存放处
  // 颜色名称，可添加，待设置rgb类型设置颜色
  color_name = {"yellow","red","green","black","blue","white","pink","purple","skyblue","lightgreen","orange",
  };

  //pub_thread.join();
  // Construct and lay out labels and slider controls.
  // Line Thickness
  QLabel* thickness_label = new QLabel( "Line Thickness" );
  QSlider* thickness_slider = new QSlider( Qt::Horizontal );
  thickness_slider->setMinimum( 1 );
  thickness_slider->setMaximum( 100 );
  // Cell Size
  QLabel* cell_size_label = new QLabel( "Cell Size" );
  QSlider* cell_size_slider = new QSlider( Qt::Horizontal );
  cell_size_slider->setMinimum( 1 );
  cell_size_slider->setMaximum( 100 );
  // Topic_Cloud
  // QLabel* Topic=new QLabel("TOPIC:");
  // QLineEdit* Topic_text=new QLineEdit();

  // Topic_Laser 输入雷达
  QLabel* Topic_laser = new QLabel("TOPIC_Laser:");
  QLineEdit* Topic_laser_text = new QLineEdit();

  // Laser
  QLabel* Size=new QLabel("Laser Size");
  QSpinBox* size=new QSpinBox();
  // 网格结构
  QGridLayout* controls_layout = new QGridLayout(); 
  controls_layout->addWidget( thickness_label, 0, 0 );
  controls_layout->addWidget( thickness_slider, 1, 0 );
  controls_layout->addWidget( cell_size_label, 2, 0 );
  controls_layout->addWidget( cell_size_slider, 3, 0 );
  // controls_layout->addWidget(Topic,0,4);
  // controls_layout->addWidget(Topic_text,0,5);

  controls_layout->addWidget(Topic_laser, 4, 0);
  controls_layout->addWidget(Topic_laser_text, 5, 0);
 
  controls_layout->addWidget(Size,6,0);
  controls_layout->addWidget(size,7,0);

  // 颜色选择QComBox
  QLabel* label_color = new QLabel("Color:");
  QComboBox * comboBox = new QComboBox();
  for(int i =0 ; i < color_name.size(); ++i){
      comboBox->addItem(color_name[i]);
  }
  controls_layout->addWidget(label_color,8,0);
  controls_layout->addWidget(comboBox,8,1);

  // Tree
  Tree_Display(controls_layout, 9);

  // 设置网格某一行伸缩比例
  controls_layout->setRowStretch(9, 6);
  // controls_layout->setRowStretch(3, 0);

  // Construct and lay out render panel.
  render_panel_ = new rviz::RenderPanel();  // RVIZ在QT上显示的类
  QHBoxLayout* main_layout = new QHBoxLayout;  // 主要layout QV垂直，QH水平
  main_layout->addLayout( controls_layout );
  main_layout->addWidget( render_panel_,2); // 存放rviz的位置, stretch 拉伸系数

  // QSplitter *splitter = new QSplitter();
  // main_layout->addWidget(splitter);
  // splitter->addWidget(render_panel_);

  // Set the top-level layout for this MyViz widget.
  setLayout( main_layout );

  // Make signal/slot connections. 进行信号/插槽连接
  connect( thickness_slider, SIGNAL( valueChanged( int )), this, SLOT( setThickness( int )));
  connect( cell_size_slider, SIGNAL( valueChanged( int )), this, SLOT( setCellSize( int )));
  // 增加雷达话题
  connect(Topic_laser_text, SIGNAL(textChanged(const QString &)), this, SLOT(setLaserTopic(const QString &)));
  connect(size,SIGNAL(valueChanged(int)),this,SLOT(setLaserSize(int)));

  // 增加颜色变换
  connect(comboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(LaserColorChanged(int)));

  // Next we initialize the main RViz classes.
  //
  // The VisualizationManager is the container for Display objects,
  // holds the main Ogre scene, holds the ViewController, etc.  It is
  // very central and we will probably need one in every usage of
  // librviz.
  manager_ = new rviz::VisualizationManager( render_panel_ ); // 是实现rviz功能的类
  // 初始化camera 这行代码实现放大 缩小 平移等操作
  render_panel_->initialize( manager_->getSceneManager(), manager_ );
  // manager_->setFixedFrame("world");
  manager_->initialize();
  manager_->startUpdate();

  // Create a Grid display.
  grid_ = manager_->createDisplay( "rviz/Grid", "adjustable grid", true );
  ROS_ASSERT( grid_ != NULL );
  
  // cloud = manager_->createDisplay( "rviz/PointCloud2", "point cloud", true );
  // ROS_ASSERT( cloud != NULL );

  laser = manager_->createDisplay( "rviz/LaserScan", "Qlaser", true );
  ROS_ASSERT( laser );
  // 新增雷达
  laser->subProp("Topic")->setValue("/scan");
  laser->subProp("Size (Pixels)")->setValue("2");
  laser->subProp("Color Transformer")->setValue("FlatColor");
  laser->subProp("Color")->setValue(comboBox->currentText());
  // cloud->subProp("Topic")->setValue("/pose_graph/octree");
  // cloud->subProp("Style")->setValue("Points");
  // cloud->subProp("Size (Pixels)")->setValue("2");
  // cloud->subProp("Color Transformer")->setValue("Intensity");
  // cloud->subProp("Invert Rainbow")->setValue("true");
  // cloud->subProp("Decay Time")->setValue("1");
  // Configure the GridDisplay the way we like it.
  grid_->subProp( "Line Style" )->setValue( "Billboards" );
  grid_->subProp( "Color" )->setValue( QColor( Qt::white ) );

  // Initialize the slider values.
  thickness_slider->setValue( 5 );
  cell_size_slider->setValue( 6 );
}

void MyViz::Tree_Display(QGridLayout* controls_layout, int index){
  QTreeWidget *menu = new QTreeWidget();
  // menu->setColumnCount(2);
  // 列头自适应大小
  menu->header()->setSectionResizeMode(QHeaderView::ResizeToContents);

  menu->setHeaderLabels(QStringList()<<"key"<<"value");
  // menu->setColumnCount(2);

  //设置不同层次菜单的缩进
  // menu->setIndentation(10);

  // global-一层树
  QTreeWidgetItem *global= new QTreeWidgetItem(menu, QStringList("全局变量"));
  menu->addTopLevelItem(global);
  global->setExpanded(true);

  //fix frame
  QTreeWidgetItem *fixed_frame = new QTreeWidgetItem(global, QStringList("Fixed Frame"));

  QLineEdit* fixed_frame_text = new QLineEdit();
  fixed_frame_text->setText("world");
  fixed_frame_text->setStyleSheet("background:transparent;border-width:0;border-style:inset");
  // 设置事件
  connect(fixed_frame_text,SIGNAL(textChanged(const QString&)),this,SLOT(setFixedFrame(const QString&)));

  // 将其他控件(非TreeWidgetItem)放入treeItem中
  menu->setItemWidget(fixed_frame, 1, fixed_frame_text); // 将container 放到 Item ,数字为列数
  

  QTreeWidgetItem *bg_color = new QTreeWidgetItem(global, QStringList("BackGround Color"));
  QLineEdit* bg_color_text = new QLineEdit();
  bg_color_text->setText("48;48;48");
  bg_color_text->setStyleSheet("border-width:0;border-style:outset");
  // 设置事件
  connect(bg_color_text,SIGNAL(textChanged(const QString&)),this,SLOT(setBackgroundColor(const QString&)));

  menu->setItemWidget(bg_color, 1, bg_color_text); // 将container 放到 Item ,数字为列数
  
  // global->addChild(fixed_frame);
  // global->setTextAlignment(1, 2); // (列数，对齐方式)
  // grid - 一层树
  QTreeWidgetItem * grid = new QTreeWidgetItem(menu, QStringList("Grid"));
  // 二层
  QTreeWidgetItem *grid_color = new QTreeWidgetItem(grid, QStringList("Color"));
  QComboBox * grid_color_cbox = new QComboBox();
  for(int i =0 ; i < color_name.size(); ++i){
      grid_color_cbox->addItem(color_name[i]);
  }
  connect(grid_color_cbox, SIGNAL(currentIndexChanged(int)), this, SLOT(GridColorChanged(int)));
  menu->setItemWidget(grid_color, 1, grid_color_cbox);

  QTreeWidgetItem *item3 = new QTreeWidgetItem(menu, QStringList("Laser"));

  QTreeWidgetItem *item13 = new QTreeWidgetItem(global, QStringList("Frame rate"));
  
  QTreeWidgetItem *item21 = new QTreeWidgetItem(grid, QStringList("Cell Size"));
  QTreeWidgetItem *item22 = new QTreeWidgetItem(grid, QStringList("Line Style"));
  QTreeWidgetItem *item23 = new QTreeWidgetItem(grid, QStringList("Color"));
  controls_layout->addWidget(menu, index, 0 );
}

void MyViz::setBackgroundColor(const QString& color_string){
  if(manager_!=NULL){
    const QString t = color_string;
    QColor new_color = parseColor(t);
    if(new_color.isValid()){
      // rviz源码中修改背景颜色的方式
      render_panel_->setBackgroundColor(qtToOgre(new_color));
      manager_->queueRender();
    }
    else
    {
      ROS_INFO_STREAM("Color invalid !!!");
    }    
  }else{
    ROS_INFO_STREAM("manager_ failed !!!");
  }
}


void MyViz::LaserColorChanged(int index){
  if(laser!=NULL) laser->subProp("Color")->setValue(color_name[index]);
  
}

void MyViz::GridColorChanged(int index){
  if(grid_!=NULL) grid_->subProp("Color")->setValue(color_name[index]);
}

// Destructor.
MyViz::~MyViz()
{
  delete manager_;
}

// This function is a Qt slot connected to a QSlider's valueChanged()
// signal.  It sets the line thickness of the grid by changing the
// grid's "Line Width" property.
void MyViz::setThickness( int thickness_percent )
{
  if( grid_ != NULL )
  {
    grid_->subProp( "Line Style" )->subProp( "Line Width" )->setValue( thickness_percent / 100.0f );
  }
}

// This function is a Qt slot connected to a QSlider's valueChanged()
// signal.  It sets the cell size of the grid by changing the grid's
// "Cell Size" Property.
void MyViz::setCellSize( int cell_size_percent )
{
  if( grid_ != NULL )
  {
    grid_->subProp( "Cell Size" )->setValue( cell_size_percent / 10.0f );
  }
}

void MyViz::setCloudTopic(const QString &newTopic){
  if(cloud!=NULL){
    cloud->subProp( "Topic" )->setValue(newTopic);
    //ROS_INFO_STREAM("cloud topic changed to => "<<newTopic.toStdString());
  }
}

void MyViz::setLaserTopic(const QString &newTopic){
  if(laser!=NULL){
    laser->subProp( "Topic" )->setValue(newTopic);
    ROS_INFO_STREAM("laser topic changed to => "<<newTopic.toStdString());
  }
  laser->setEnabled(true);
  // laser->subProp( "Color" )->setValue( QColor( Qt::yellow ) );
  manager_->setFixedFrame("laser");
  manager_->startUpdate();
}

void MyViz::setFixedFrame(const QString & fix_frame){
  manager_->setFixedFrame(fix_frame);
  manager_->startUpdate();
}

void MyViz::setCloudSize(int cloudsize){
  if(cloud!=NULL){
    cloud->subProp("Size (Pixels)")->setValue(cloudsize);
  }
}

void MyViz::setLaserSize(int lasersize){
  if(laser!=NULL){
    laser->subProp("Size (Pixels)")->setValue(lasersize);
  }
}

void MyViz::subCallback(const std_msgs::String& msg){
  ROS_INFO_STREAM("receive message!");
}

void MyViz::pubThread(){
  while(ros::ok()){
    ROS_INFO_STREAM_ONCE("here is in publish process!");
  }
}